(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* The Sum Type as Error Monad *}

theory Error_Monad
imports "~~/src/HOL/Library/Monad_Syntax"
  "$AFP/Matrix/Utility"
begin

text {*The \emph{monadic bind} operation.*}
definition 
  bind :: "'e + 'a \<Rightarrow> ('a \<Rightarrow> 'e + 'b) \<Rightarrow> 'e + 'b"
where
  "bind m f = (case m of Inr x \<Rightarrow> f x | Inl e \<Rightarrow> Inl e)"

adhoc_overloading
  Monad_Syntax.bind bind

abbreviation (input) return where "return \<equiv> Inr"
abbreviation (input) error where "error \<equiv> Inl"
abbreviation (input) run where "run \<equiv> projr"


subsection {* Monad Laws *}

lemma return_bind:
  "(Inr x \<guillemotright>= f) = f x" by (simp add: bind_def)

lemma bind_return:
  "(m \<guillemotright>= Inr) = m" by (cases m) (simp_all add: bind_def)

lemma error_bind:
  "(error e \<guillemotright>= f) = error e" by (simp add: bind_def)

lemma bind_assoc:
  fixes m :: "'a + 'b"
  shows "((m \<guillemotright>= f) \<guillemotright>= g) = (m \<guillemotright>= (\<lambda>x. f x \<guillemotright>= g))"
  by (cases m) (simp_all add: bind_def)

lemma bind_cong [fundef_cong]:
  fixes m1 m2 :: "'e + 'a"
    and f1 f2 :: "'a \<Rightarrow> 'e + 'b"
  assumes "m1 = m2"
    and "\<And>y. m2 = Inr y \<Longrightarrow> f1 y = f2 y"
  shows "(m1 \<guillemotright>= f1) = (m2 \<guillemotright>= f2)"
  using assms by (cases "m1") (auto simp: bind_def)

definition
  catch :: "'e + 'a \<Rightarrow> ('e \<Rightarrow> 'f + 'a) \<Rightarrow> 'f + 'a"
    ("(try(/ _)/ catch(/ _))" [12, 12] 13)
where
  "(try m catch f) = (case m of Inl e \<Rightarrow> f e | Inr x \<Rightarrow> Inr x)"

lemma catch_splits:
  "P (try m catch f) \<longleftrightarrow>
    (\<forall>e. m = Inl e \<longrightarrow> P (f e)) \<and>
    (\<forall>x. m = Inr x \<longrightarrow> P (Inr x))"
  "P (try m catch f) \<longleftrightarrow>
    (\<not> ((\<exists>e. m = Inl e \<and> \<not> P (f e)) \<or>
        (\<exists>x. m = Inr x \<and> \<not> P (Inr x))))"
  by (case_tac [!] m) (simp_all add: catch_def)

abbreviation
  update_error :: "'e + 'a \<Rightarrow> ('e \<Rightarrow> 'f) \<Rightarrow> 'f + 'a" (infixl "<+?" 61)
where
  "m <+? f \<equiv> try m catch (\<lambda>x. error (f x))"

abbreviation
  replace_error :: "'e + 'a \<Rightarrow> 'e \<Rightarrow> 'e + 'a" (infixl "<?" 61)
where
  "m <? e \<equiv> m <+? (\<lambda>x. e)"

lemma catch_return:
  "(try return x catch f) = Inr x" by (simp add: catch_def)

lemma catch_error:
  "(try error e catch f) = f e" by (simp add: catch_def)

lemma update_error_return:
  "(m <+? c = return x) = (m = return x)"
  by (cases m) (auto simp: catch_error catch_return)

definition isOK :: "'e + 'a \<Rightarrow> bool" where
  "isOK m \<longleftrightarrow> (case m of Inl e \<Rightarrow> False | Inr x \<Rightarrow> True)"

lemma isOK_E [elim]:
  "\<lbrakk>isOK m; \<And>x. m = Inr x \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
  unfolding isOK_def by (cases m) simp_all

lemma isOK_I [intro]:
  assumes "m = Inr x" shows "isOK m"
  using assms by (simp add: isOK_def)

lemma isOK_return:
  "isOK (Inr x) \<longleftrightarrow> True" by (simp add: isOK_def)

lemma isOK_error:
  "isOK (error e) \<longleftrightarrow> False" by (simp add: isOK_def)

lemma isOK_bind:
  "isOK (m \<guillemotright>= f) \<longleftrightarrow> isOK m \<and> isOK (f (run m))"
  by (cases m) (simp_all add: error_bind isOK_error return_bind isOK_return)

lemma isOK_update_error:
  "isOK (m <+? f) \<longleftrightarrow> isOK m"
  by (cases m) (simp_all add: catch_error isOK_error catch_return isOK_return)

lemma case_ok_pair[simp]: "isOK(case lr of (l,r) \<Rightarrow> P l r) = (case lr of (l,r) \<Rightarrow> isOK(P l r))" by auto
lemma case_ok_option[simp]: "isOK(case opt of None \<Rightarrow> P | Some v \<Rightarrow> Q v) = (case opt of None \<Rightarrow> isOK P | Some v \<Rightarrow> isOK(Q v))" by (cases opt, auto)        

lemma run_bind:
  assumes "isOK m" shows "run (m \<guillemotright>= f) = run (f (run m))"
  using assms by (rule isOK_E) (simp add: return_bind)

lemma run_catch:
  assumes "isOK m" shows "run (try m catch f) = run m"
  using assms by (rule isOK_E) (simp add: catch_return)

lemmas run_simps = run_bind run_catch

fun
  foldM :: "('a \<Rightarrow> 'b \<Rightarrow> 'e + 'a) \<Rightarrow> 'a \<Rightarrow> 'b list \<Rightarrow> 'e + 'a"
where 
  "foldM f d [] = return d" |
  "foldM f d (x # xs) = do { y \<leftarrow> f d x; foldM f y xs }"

fun
  forallM_index_aux :: "('a \<Rightarrow> nat \<Rightarrow> 'e + unit) \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> (('a \<times> nat) \<times> 'e) + unit"
where
  "forallM_index_aux P i [] = return ()" |
  "forallM_index_aux P i (x # xs) = do {
     P x i <+? Pair (x, i);
     forallM_index_aux P (Suc i) xs
   }"

lemma isOK_forallM_index_aux:
  "isOK (forallM_index_aux P n xs) = (\<forall>i<length xs. isOK (P (xs ! i) (i + n)))"
proof (induct xs arbitrary: n)
  case Nil then show ?case by (simp add: isOK_return)
next
  case (Cons x xs)
  have "(\<forall>i<length (x # xs). isOK (P ((x # xs) ! i) (i + n))) \<longleftrightarrow>
    (isOK (P x n) \<and> (\<forall>i<length xs. isOK (P (xs ! i) (i + Suc n))))"
    by (simp add: all_Suc_conv)
  then show ?case
    unfolding Cons [of "Suc n", symmetric] by (simp add: isOK_return isOK_update_error isOK_bind)
qed

definition
  forallM_index :: "('a \<Rightarrow> nat \<Rightarrow> 'e + unit) \<Rightarrow> 'a list \<Rightarrow> (('a \<times> nat) \<times> 'e) + unit"
where
  "forallM_index P xs = forallM_index_aux P 0 xs"

lemma forallM_index [fundef_cong]:
  fixes c :: "'a \<Rightarrow> nat \<Rightarrow> 'e + unit"
  assumes "\<And>x i. x \<in> set xs \<Longrightarrow> c x i = d x i"
  shows "forallM_index c xs = forallM_index d xs"
proof -
  {
    fix n
    from assms have "forallM_index_aux c n xs = forallM_index_aux d n xs"
      by (induct xs arbitrary: n) simp_all
  }
  then show ?thesis unfolding forallM_index_def by simp
qed

hide_const forallM_index_aux

lemma isOK_forallM_index:
  "isOK (forallM_index P xs) \<longleftrightarrow> (\<forall>i<length xs. isOK (P (xs ! i) i))"
  unfolding forallM_index_def isOK_forallM_index_aux by simp

text {*Check whether @{term f} succeeds for all elements of a given list. In case it doesn't,
return the first offending element together with the produced error.*}
fun
  forallM :: "('a \<Rightarrow> 'e + unit) \<Rightarrow> 'a list \<Rightarrow> ('a * 'e) + unit"
where
  "forallM f [] = return ()" |
  "forallM f (x # xs) = (f x <+? Pair x \<guillemotright> forallM f xs)"

lemma isOK_forallM:
  "isOK (forallM f xs) \<longleftrightarrow> (\<forall>x\<in>set xs. isOK (f x))"
  by (induct xs) (simp_all add: isOK_return isOK_update_error isOK_bind)

text {*Check whether @{term f} succeeds for at least one element of a given list.
In case it doesn't, return the list of produced errors.*}
fun
  existsM :: "('a \<Rightarrow> 'e + unit) \<Rightarrow> 'a list \<Rightarrow> 'e list + unit"
where
  "existsM f [] = error []" |
  "existsM f (x # xs) = (try (f x) catch (\<lambda>e. (existsM f xs) <+? Cons e))"

lemma isOK_existsM:
  "isOK (existsM f xs) \<longleftrightarrow> (\<exists>x\<in>set xs. isOK (f x))"
proof (induct xs)
  case (Cons x xs)
  show ?case
  proof (cases "f x")
    case (Inl e)
    with Cons show ?thesis by (auto simp add: catch_error isOK_update_error)
  qed (auto simp add: catch_def)
qed (auto simp add: isOK_error)

lemma isOK_if_then:
  "isOK (if b then m else error e) \<longleftrightarrow> b \<and> isOK m"
  by (simp add: isOK_error)

lemma isOK_if_else:
  "isOK (if b then error e else m) \<longleftrightarrow> \<not> b \<and> isOK m"
  by (simp add: isOK_error)

lemma isOK_if_then': "isOK(if p then Inr r else b) = (p \<or> isOK(b))"
  by (simp add: isOK_return)

lemma isOK_if_else': "isOK(if p then b else Inr r) = (\<not>p \<or> isOK(b))"
  by (simp add: isOK_return)

lemma isOK_if_then_else: "isOK(if p then a else b) = (p \<and> isOK a \<or> \<not> p \<and> isOK b)"
  by simp

lemmas isOK_simps [simp] =
  isOK_error isOK_bind isOK_return isOK_update_error
  isOK_forallM_index
  isOK_forallM
  isOK_existsM
  isOK_if_then isOK_if_else
  isOK_if_then' isOK_if_else'

fun
  sequence :: "('e + 'a) list \<Rightarrow> 'e + 'a list"
where
  "sequence [] = Inr []" |
  "sequence (m # ms) = do {
    x \<leftarrow> m;
    xs \<leftarrow> sequence ms;
    return (x # xs)
  }"

lemmas monad_simps [simp] =
  return_bind bind_return error_bind bind_assoc
  catch_return catch_error
  run_simps update_error_return


subsection {* Monadic Map for Error Monad *}

fun mapM :: "('a \<Rightarrow> 'e + 'b) \<Rightarrow> 'a list \<Rightarrow> 'e + 'b list" where
  "mapM f [] = Inr []" |
  "mapM f (x#xs) = do {
    y \<leftarrow> f x;
    ys \<leftarrow> mapM f xs;
    Inr (y # ys)
  }"

lemma mapM_Inl:
  "(\<exists>e. mapM f xs = Inl e) \<longleftrightarrow> (\<exists>x\<in>set xs. \<exists>e. f x = Inl e)"
proof (induct xs)
  case (Cons x xs)
  then show ?case
    by (cases "f x", simp_all, cases "mapM f xs", simp_all)
qed simp

lemma mapM_Inr:
  assumes "mapM f xs = Inr ys"
  shows "ys = map (\<lambda>x. projr (f x)) xs \<and> (\<forall>x\<in>set xs. \<forall>e. f x \<noteq> Inl e)"
using assms
proof (induct xs arbitrary: ys)
  case (Cons x xs ys)   
  then show ?case
    by (cases "f x", simp, cases "mapM f xs", auto)
qed simp

lemma mapM_Inr_idx:
  assumes *: "mapM f xs = Inr ys" and "i < length xs" 
  shows "\<exists>y. f (xs ! i) = Inr y \<and> ys ! i = y"
proof -
  note ** = mapM_Inr [OF *, unfolded set_conv_nth]
  with assms have "\<And>e. f (xs ! i) \<noteq> Inl e" by auto
  then obtain y where "f (xs ! i) = Inr y" by (cases "f (xs ! i)") auto
  then have "f (xs ! i) = Inr y \<and> ys ! i = y" unfolding ** [THEN conjunct1] using assms by auto
  then show ?thesis ..
qed

lemma mapM_cong [fundef_cong]:
  assumes "xs = ys" and "\<And>x. x \<in> set ys \<Longrightarrow> f x = g x"
  shows "mapM f xs = mapM g ys"
  unfolding assms(1) using assms(2) by (induct ys) auto

lemma bindE [elim]:
  assumes "(p \<guillemotright>= f) = Inr x"
  obtains y where "p = Inr y" and "f y = Inr x"
  using assms by (cases p) simp_all

end
