(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
(*
  Call trees
*)

signature CALL_TREE =
sig
  datatype app_type = Rec | Constr | Def

  datatype tree =
    Variable of term * string
  | Case of term * Datatype.info * string * string list * tree * (string list * term * tree) list
  | If of term * string * string list * tree * tree * tree
  | App of term * app_type * string * string * tree list
  | Unsupported of term * string * tree list

  val i_term : tree -> term
  val i_term_dps : tree -> (term * tree) list

  val add_defined_syms : tree -> string list -> string list
  val add_case_syms : tree -> string list -> string list
  val add_sim_syms : string -> tree -> string list -> string list

  val analyze_rules : string list -> (string -> bool) -> (string -> bool) -> theory -> term list -> 
    (string * (term * term * tree) list) list
  (* (term * term * tree) list: lhs, lhs#, rhs as call-tree *)

end

structure Call_Tree : CALL_TREE =
struct

open Isafor_Basics

(* explicit trees *)

datatype app_type = Rec | Constr | Def

datatype tree =
  Variable of term * string
| Case of term * Datatype.info * string * string list * tree * (string list * term * tree) list
| If of term * string * string list * tree * tree * tree
| App of term * app_type * string * string * tree list (* fst string: name, snd string: fresh_var *)
| Unsupported of term * string * tree list (* string is name for fresh variable, tree list may contain recursive calls *)


fun fold_pats thy f t =
  let 
    val (Const (_, T), ts) = strip_comb t
    val elimT as (Type (dt, _)) = List.last (binder_types T)
    val constrs = the (Datatype.get_constrs thy dt)
    
    fun fold_pat ((cname, cT), lamterm) =
      let
        val n = length (binder_types cT)
        val (vs', b) = Term.strip_abs_eta n lamterm
        val pat = 
          list_comb (Const (cname, map snd vs' ---> elimT), map Free vs')
      in
        f (map fst vs') pat b
      end

    val (lams, _) = split_last ts
  in
    fold fold_pat (constrs ~~ lams)
  end



fun mk_tree _ _ _ _ _ _ (t as Free (v, _)) x = (Variable (t, v), x)
  | mk_tree f fs is_constr is_sim thy vs t i =
  let
    val (c, ts) = strip_comb t
    fun new_name f i = f ^ "__" ^ string_of_int i
    val fresh_var = mk_fresh_var_name
    fun extra_vars ts = fold Term.add_frees ts [] |> rev |> map fst
    val is_first_order = case fastype_of t of
       Type (n,_) => n <> "fun"
     | _ => true
  in
    if is_Const c andalso is_first_order then case c of 
     Const (n, _) =>
    if n = @{const_name HOL.If} then
      let
        val [cond,thn,els] = ts
        val (cond_tree, i') = mk_tree f fs is_constr is_sim thy vs cond i
        val xs = extra_vars [thn, els]
        val (thn_tree, i'') = mk_tree f fs is_constr is_sim thy xs thn (i' + 1)
        val (els_tree, i''') = mk_tree f fs is_constr is_sim thy xs els i''
      in
        (If (t, new_name f i', xs, cond_tree, thn_tree, els_tree), i''')
      end
    else case Datatype.info_of_case thy n of
      NONE => 
        let val r = member (op =) fs n
            val (ts', i') = fold_map (mk_tree f fs is_constr is_sim thy vs) ts i
            val fresh = fresh_var i'
            val i'' = i' + 1
        in if r orelse is_sim n then
           let val app_type =
                    if r then Rec else if is_constr n then Constr else Def
           in (App (t, app_type, n, fresh, ts'), i'') end
           else (Unsupported (t,fresh,ts'), i'')
        end 
    | SOME dt =>
         let
           val (lams, arg) = split_last ts
           val (arg_tree, i') = mk_tree f fs is_constr is_sim thy vs arg i

           val xs = extra_vars lams

           fun add_subtree vs' pat b (i, sts) =
             let
               val (b_tree, i') = 
                 mk_tree f fs is_constr is_sim thy (vs' @ xs) b i
             in
               (i', (vs', pat, b_tree) :: sts)
             end
           val (i'', sts) = fold_pats thy add_subtree t (i' + 1, [])
          in
            (Case (t, dt, new_name f i', xs, arg_tree, rev sts), i'')
          end      
      else 
      let val (ts', i') = fold_map (mk_tree f fs is_constr is_sim thy vs) ts i
         in (Unsupported (t,fresh_var i', ts'),i' + 1)
      end
    end

fun ap_type_con Constr = true
  | ap_type_con _ = false 

fun i_term (Case (_, _, f, extra, arg, _)) =  mk_app false true f (i_term arg :: map mk_var extra)
  | i_term (Variable (_, x)) = mk_var x
  | i_term (App (_, at, f, _, ts)) = mk_app (ap_type_con at) true f (map i_term ts)
  | i_term (If (_, f, extra, cond, _, _)) = mk_app false true f (i_term cond :: map mk_var extra)
  | i_term (Unsupported (_, x, _)) = mk_var x

fun i_term_below (Case (_, _, f, extra, arg, _)) =  mk_app false true f (i_term_below arg :: map mk_var extra)
  | i_term_below (Variable (_, x)) = mk_var x
  | i_term_below (App (_, at, f, x, ts)) = if at = Rec then mk_var x else mk_app (ap_type_con at) true f (map i_term_below ts)
  | i_term_below (If (_, f, extra, cond, _, _)) = mk_app false true f (i_term_below cond :: map mk_var extra)
  | i_term_below (Unsupported (_, x, _)) = mk_var x

fun i_term_dps (t as Case (_, _, f, extra, arg, _)) = (mk_sharp f (i_term arg :: map mk_var extra),t)
        :: i_term_dps arg
  | i_term_dps (Variable _) = []
  | i_term_dps (Unsupported (_,_,ts)) = maps i_term_dps ts
  | i_term_dps (t as App (_, ap, f, _, ts)) = 
      (if ap = Rec then [(mk_sharp f (map i_term_below ts),t)] else []) @ 
     maps i_term_dps ts
  | i_term_dps (t as If (_, f, extra, cond, _, _)) = (mk_sharp f (i_term cond :: map mk_var extra),t) :: i_term_dps cond


(* FIXME: we do not need all defined symbols; only those that occur in arguments
   of recursive calls *)
fun add_defined_syms (Case (_, _, _, _, arg, children)) =
      add_defined_syms arg #> fold (add_defined_syms o #3) children
  | add_defined_syms (Variable _) = I
  | add_defined_syms (App (_, Def, f, _, ts)) =
      insert (op =) f #> fold add_defined_syms ts
  | add_defined_syms (App (_, _, _, _, ts)) =
      fold add_defined_syms ts
  | add_defined_syms (If (_, _, _, cond, thn, els)) =
      fold add_defined_syms [cond, thn, els]
  | add_defined_syms (Unsupported _) = I

fun contains_sym f (Case (_, _, _, _, arg, cs)) =
      contains_sym f arg orelse exists (contains_sym f o #3) cs
  | contains_sym _ (Variable _) = false
  | contains_sym _ (Unsupported _) = false
  | contains_sym f (App (_, _, g, _, ts)) = f = g orelse exists (contains_sym f) ts
  | contains_sym f (If (_, _, _, c, t, e)) = exists (contains_sym f) [c,t,e]

fun add_sim_syms f =
  let
    fun add b (Case (_, _, _, _, arg, cs)) =
          add (b orelse exists (contains_sym f o #3) cs) arg #> fold (add b o #3) cs
      | add _ (Variable _) = I
      | add _ (Unsupported _) = I
      | add b (App (_, Def, g, _, ts)) =
          (if b then insert (op =) g else I) #> fold (add b) ts
      | add b (App (_, apt, _, _, ts)) = fold (add (b orelse apt = Rec)) ts
      | add b (If (_, _, _, c, t, e)) =
          add (b orelse exists (contains_sym f) [t,e]) c #> fold (add b) [t,e]
  in add false end

fun add_case_syms (Case (_, _, f, _, arg, children)) =
      insert (op =) f #> add_case_syms arg #> fold (add_case_syms o #3) children
  | add_case_syms (Variable _) = I
  | add_case_syms (Unsupported _) = I
  | add_case_syms (App (_, _, _, _, ts)) =
      fold add_case_syms ts
  | add_case_syms (If (_, f, _, cond, thn, els)) =
      insert (op =) f #> fold add_case_syms [cond, thn, els]

fun analyze_rules fs is_constr is_sim thy eqs =
  let
    fun case_rules (Case (_, _, f, extra, arg, children)) =
        (f, map (fn (_, p, tr) => 
        let val args = mk_term is_constr p :: map mk_var extra
            val lhs = mk_app false true f args
            val lhs_sharp = mk_sharp f args
        in (lhs, lhs_sharp,tr) end) children)
      :: case_rules arg @ maps (case_rules o #3) children
      | case_rules (Variable _) = []
      | case_rules (Unsupported _) = []
      | case_rules (App (_, _, _, _, trs)) = maps case_rules trs
      | case_rules (If (_, f, extra, cond, thn, els)) =
        let val targs = mk_app true true @{const_name HOL.True} [] :: map mk_var extra
            val fargs = mk_app true true @{const_name HOL.False} [] :: map mk_var extra
            
        in (f, [(mk_app false true f targs, mk_sharp f targs, thn),
             (mk_app false true f fargs, mk_sharp f fargs, els)])
          :: maps case_rules [cond, thn, els] end
    fun mk_rule eq i =
      let
        val (lhs, rhs) = HOLogic.dest_eq (HOLogic.dest_Trueprop eq)
        val vs = Term.add_frees lhs [] |> rev |> map fst
        val Const (f, _) = head_of lhs
        val (tr, i') = mk_tree f fs is_constr is_sim thy vs rhs i
       in ((f, (mk_term is_constr lhs, mk_sharp_lhs is_constr lhs, tr)), i') end
    val main_rules_trees = fst (fold_map mk_rule eqs 0)
    val trees = map (#3 o snd) main_rules_trees
    val main_rules = AList.group (op =) main_rules_trees
    val more_rules = maps case_rules trees
  in
    main_rules @ more_rules
  end

end
