(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

signature TRS_DATA =
sig
  val get_def : theory -> string -> (string * thm) option
  val get_def_dp : theory -> string -> (string * thm * string * thm) option
  val get_trees : theory -> string -> (term * Call_Tree.tree) list
  val get_dp_trees : theory -> string -> (term * Call_Tree.tree) list
  val get_rules : theory -> string -> thm list
  val get_dp_rules : theory -> string -> (thm * Call_Tree.tree) list
  val get_incl : theory -> string * string -> thm option
  val is_constr : theory -> string -> bool
  val is_sim : theory -> string -> bool
  val add_constrs : string list -> theory -> theory

  val add_sim : thm -> Context.generic -> Context.generic
  val get_sims : Proof.context -> thm list
  val get_embs : Proof.context -> thm list
  val sim_for_trs : Proof.context -> string -> string -> thm

  val add_emb : attribute

  val define_trs : thm list -> theory -> theory
  val define_dps : thm list -> theory -> theory
         (* P, R, call-trees *)
  val declare_constrs : string -> theory -> theory
  val all_trs_defs : Proof.context -> thm list

  val setup : theory -> theory

  structure Embedding : NAMED_THMS
end

structure Trs_Data (*: TRS_DATA*) =
struct

fun duplicate_sim_err const_name =
  error ("Constant " ^ quote const_name ^ " does already have a simulation lemma");

open Isafor_Basics

structure Simulation = Generic_Data
(
  type T = thm Symtab.table
  val empty = Symtab.empty
  val extend = I
  val merge =
    Symtab.merge (K true)
    handle Symtab.DUP c => duplicate_sim_err c; (* FIXME? *)
)

fun const_for_sim thm =
  let val ((l,_),_) = dest_inrel (prop_of thm)
  in (drop_prefix o fst) (dest_app l) end

fun add_sim thm = Simulation.map (Symtab.update_new (const_for_sim thm, thm))
val get_sims = map snd o Symtab.dest o Simulation.get o Context.Proof

val get_sim_for_const = Symtab.lookup o Simulation.get o Context.Proof
fun the_sim_for_const ctxt name =
  case get_sim_for_const ctxt name of
    SOME thm => thm
  | NONE => error ("No simulation lemma for " ^ quote name)

structure Embedding = 
  Named_Thms(val name = @{binding emb}  val description = "emb rules");

structure Subst = 
  Named_Thms(val name = @{binding trs_subst}  val description = "substitution rules for trs representation");

val get_embs = Embedding.get
val add_emb = Embedding.add

structure TrsData = Theory_Data(
  type T = (
    (string * thm) Symtab.table  (* deftab *)
  * (string * thm * string * thm) Symtab.table  (* dp-deftab *)
  * ((term * Call_Tree.tree) list) Symtab.table (* treetab *)
  * ((term * Call_Tree.tree) list) Symtab.table (* dp-treetab *)
  * (thm list) Symtab.table (* ruletab *)
  * ((thm * Call_Tree.tree) list) Symtab.table (* dp-ruletab *)
  * thm Symreltab.table (* trs hierarchy inclusions *)
  * unit Symtab.table (* constructors *)
  )
  val empty = (Symtab.empty, Symtab.empty, Symtab.empty, Symtab.empty, Symtab.empty, Symtab.empty, Symreltab.empty, Symtab.empty)
  val extend = I
  fun merge ((x1,x2,x3,x4,x5,x6,x7,x8), (y1,y2,y3,y4,y5,y6,y7,y8)) = (
    Symtab.merge (K true) (x1,y1),
    Symtab.merge (K true) (x2,y2),
    Symtab.merge (K true) (x3,y3),
    Symtab.merge (K true) (x4,y4),
    Symtab.merge (K true) (x5,y5),
    Symtab.merge (K true) (x6,y6),
    Symreltab.merge (K true) (x7,y7),
    Symtab.merge (K true) (x8,y8))
)

val get_deftab = #1 o TrsData.get
val get_dp_deftab = #2 o TrsData.get
val get_treetab = #3 o TrsData.get
val get_dp_treetab = #4 o TrsData.get
val get_ruletab = #5 o TrsData.get
val get_dp_ruletab = #6 o TrsData.get
val get_incltab = #7 o TrsData.get
val get_constrs = #8 o TrsData.get

fun map_deftab f = TrsData.map (fn (a,y,b,x,c,z,d,e) => (f a, y, b, x, c, z, d, e))
fun map_dp_deftab f = TrsData.map (fn (a,y,b,x,c,z,d,e) => (a, f y, b, x, c, z, d, e))
fun map_treetab f = TrsData.map (fn (a,y,b,x,c,z,d,e) => (a, y, f b, x, c, z, d, e))
fun map_dp_treetab f = TrsData.map (fn (a,y,b,x,c,z,d,e) => (a, y, b, f x, c, z, d, e))
fun map_ruletab f = TrsData.map (fn (a,y,b,x,c,z,d,e) => (a, y, b, x, f c, z, d, e))
fun map_dp_ruletab f = TrsData.map (fn (a,y,b,x,c,z,d,e) => (a, y, b, x, c, f z, d, e))
fun map_incltab f = TrsData.map (fn (a,y,b,x,c,z,d,e) => (a, y, b, x, c, z, f d, e))

fun add_constrs cs =
  TrsData.map (fn (a,y,b,x,c,z,d,e) => (a, y, b, x, c, z, d, fold (Symtab.update o rpair ()) cs e))

val get_def = Symtab.lookup o get_deftab
val get_dp_def = Symtab.lookup o get_dp_deftab
val get_trees = Symtab.lookup_list o get_treetab
val get_dp_trees = Symtab.lookup_list o get_dp_treetab
fun get_rules thy name = Symtab.lookup_list (get_ruletab thy) name
fun get_dp_rules thy name = Symtab.lookup_list (get_dp_ruletab thy) name
val get_incl = Symreltab.lookup o get_incltab
val is_constr = Symtab.defined o get_constrs
val is_sim = Symtab.defined o Simulation.get o Context.Theory 

fun sim_for_trs ctxt f_name trs_name = 
  let
    val sim = the_sim_for_const ctxt (drop_prefix f_name)
  in
    case dest_inrel (prop_of sim) of
      ((_, _), _ $ (_ $ Const (trs_name', _))) =>
        let val incl = the (get_incl (Proof_Context.theory_of ctxt) (trs_name', trs_name))
        in @{thm Set.subsetD} OF [incl, sim] end
    | _ => sim (* constructor sim rule *)
  end

fun set_member_tac m i =
  if m = 0 then rtac @{thm Set.insertI1} i
  else rtac @{thm Set.insertI2} i THEN set_member_tac (m - 1) i

fun extract_dp_rule ctxt lr defthm =
  let
    open Conv
    val thy = Proof_Context.theory_of ctxt
    val (c, rhs) = Logic.dest_equals (prop_of defthm)
    val (rs :: _) = Function_Lib.dest_binop_list union_const_name rhs
    val rule = HOLogic.dest_set rs |> map_index I
      |> filter (fn (_, r) => (r = lr))
      |> hd
    fun prove_rule (m, r) =
      let 
        val cgoal = cterm_of thy (HOLogic.mk_Trueprop (HOLogic.mk_mem (r, c)))
      in
      Goal.prove_internal ctxt [] cgoal
        (K (CONVERSION (arg_conv (arg_conv (rewr_conv defthm))) 1
            THEN TRY (rtac @{thm Set.UnI1} 1)
            THEN set_member_tac m 1))
      |> subst_closure ctxt @{thm TermFun_Setup.dp_stepI}
      end
  in
    prove_rule rule
  end

fun extract_rules ctxt f defthm =
  let
    open Conv
    val thy = Proof_Context.theory_of ctxt
    val (c, rhs) = Logic.dest_equals (prop_of defthm)
    val (rs :: _) = Function_Lib.dest_binop_list union_const_name rhs
    val rules = HOLogic.dest_set rs |> map_index I
      |> filter (fn (_, r) => 
        (r |> HOLogic.dest_prod |> fst |> dest_app |> fst |> drop_prefix) = f)
    fun prove_rule (m, r) =
      let 
        val cgoal = cterm_of thy (HOLogic.mk_Trueprop (HOLogic.mk_mem (r, c)))
      in
      Goal.prove_internal ctxt [] cgoal
        (K (CONVERSION (arg_conv (arg_conv (rewr_conv defthm))) 1
            THEN TRY (rtac @{thm Set.UnI1} 1)
            THEN set_member_tac m 1))
      |> subst_closure ctxt @{thm TermFun_Setup.tf_Q_qrstepI}
      end
  in
    map prove_rule rules
  end

val Un_subsetI2 = @{lemma "A <= C ==> A <= B Un C" by auto}

fun extract_subsystems ctxt defthm =
  let
    open Conv
    val thy = Proof_Context.theory_of ctxt
    val (c, rhs) = Logic.dest_equals (prop_of defthm)
    val (_ :: subs) = Function_Lib.dest_binop_list union_const_name rhs
    fun prove_sub (m, s) =
      let 
        val cgoal = HOLogic.mk_binrel @{const_name Orderings.ord_class.less_eq}
             (mk_star (mk_qrstep mk_default_q s), mk_star (mk_qrstep mk_default_q c))
          |> HOLogic.mk_Trueprop |> cterm_of thy
      in
        ((fst (dest_Const s), fst (dest_Const c)), Goal.prove_internal ctxt [] cgoal
          (K (rtac @{thm Transitive_Closure.rtrancl_mono} 1
            THEN rtac @{thm Q_Restricted_Rewriting.qrstep_mono[OF _ subset_refl]} 1
            THEN CONVERSION (arg_conv (arg_conv (rewr_conv defthm))) 1
            THEN REPEAT_DETERM_N (m + 1) (rtac Un_subsetI2 1)
            THEN (rtac @{thm Set.Un_upper1} 1 ORELSE rtac @{thm Set.subset_refl} 1))))
      end
  in
    map_index prove_sub subs
  end

fun define_trs ceqns thy =
  let
    val eqns = map (Logic.strip_imp_concl o prop_of o Thm.unvarify_global) ceqns
    val fs = eqns
      |> map (fst o dest_Const o head_of o fst o HOLogic.dest_eq o HOLogic.dest_Trueprop)
      |> distinct (op =)
    val base_name = map Long_Name.base_name fs
      |> space_implode "_"
    val binding = Binding.name "trs"
      |> Binding.qualify true base_name

    val deftab = get_deftab thy
    fun trs_sym n = case Symtab.lookup deftab n of
      SOME (c, _) => Const (c, trsT)
    | NONE => error ("No TRS representation of " ^ quote n ^ ".")
    val new_trees = Call_Tree.analyze_rules fs (is_constr thy) (is_sim thy) thy eqns
                  |> map (apsnd (map (fn (l,_,t) => (l,t))))
    val treetab' = get_treetab thy
      |> fold Symtab.update_new new_trees
    val lhss_trees = map (Symtab.lookup_list treetab') fs
    val trees = maps (map #2) lhss_trees
    val case_syms = fold Call_Tree.add_case_syms trees [] |> rev
    val lhss_trees' = map (Symtab.lookup_list treetab') (fs @ case_syms)
    val trs = maps (map (apsnd Call_Tree.i_term)) lhss_trees'
      |> map HOLogic.mk_prod
      |> HOLogic.mk_set ruleT

    val sub_consts =
      fold Call_Tree.add_defined_syms (maps (map snd) lhss_trees') [] |> rev
      |> map trs_sym |> distinct (op aconv)
    val trs = foldr1 (HOLogic.mk_binop union_const_name)
      (trs :: sub_consts)

    val thy' = Sign.add_consts [(binding, trsT, NoSyn)] thy
    val Rf = Sign.full_name thy' binding
    val def = Logic.mk_equals (Const (Rf, trsT), trs)
    val ([defthm], thy'') = Global_Theory.add_defs false
       [((Binding.map_name (suffix "_def") binding, def), [])] thy'
    val ctxt'' = Proof_Context.init_global thy''

    val inclusions = extract_subsystems ctxt'' defthm
    fun rules f = 
      (def_prefix ^ f, (extract_rules ctxt'' f defthm))
  in
    thy''
    |> map_deftab (fold (fn f => Symtab.update (f, (Rf, defthm))) fs)
    |> map_treetab (K treetab')
    |> map_ruletab (fold Symtab.update_new (map rules (fs @ case_syms)))
    |> map_incltab (fold Symreltab.update_new inclusions)
  end

fun define_dps ceqns thy =
  let
    val eqns = map (Logic.strip_imp_concl o prop_of o Thm.unvarify_global) ceqns
    val fs = eqns
      |> map (fst o dest_Const o head_of o fst o HOLogic.dest_eq o HOLogic.dest_Trueprop)
      |> distinct (op =)

    val deftab = get_deftab thy
    fun trs_sym n = case Symtab.lookup deftab n of
      SOME (c, _) => Const (c, trsT)
    | NONE => error ("No TRS representation of " ^ quote n ^ ".")
    val new_trees = Call_Tree.analyze_rules fs (is_constr thy) (is_sim thy) thy eqns
                  |> map (apsnd (map (fn (_,l,t) => (l,t))))
    val treetab' = get_dp_treetab thy
      |> fold Symtab.update_new new_trees
    val lhss_trees = map (Symtab.lookup_list treetab') fs
    val trees = maps (map #2) lhss_trees
    val case_syms = fold Call_Tree.add_case_syms trees [] |> rev
    val f_lhss_trees' = map (fn f => (f,Symtab.lookup_list treetab' f)) (fs @ case_syms)
    val lhss_trees' = map snd f_lhss_trees'

    val dp_rules = f_lhss_trees'
                 |> map (apsnd (map (fn (l,t) => map (fn (r,t) => (HOLogic.mk_prod (l,r),t)) (Call_Tree.i_term_dps t)))) 
    val dps = maps snd dp_rules
      |> maps I
      |> map fst
      |> HOLogic.mk_set Isafor_Basics.ruleT

    val base_name = map Long_Name.base_name fs
      |> space_implode "_"

    val sub_consts_dps =
      fold (Call_Tree.add_sim_syms base_name) (maps (map snd) lhss_trees') [] |> rev
      |> map trs_sym |> distinct (op aconv)
    val trs_dps = 
        foldr1 (HOLogic.mk_binop union_const_name)
        (HOLogic.mk_set ruleT [] ::  sub_consts_dps)
    val binding = Binding.name "dps"
      |> Binding.qualify true base_name
    val binding' = Binding.name "trsdps"
      |> Binding.qualify true base_name

    val thy' = Sign.add_consts [(binding, trsT, NoSyn)] thy
    val Pf = Sign.full_name thy' binding
    val def = Logic.mk_equals (Const (Pf, trsT), dps)
    val ([defthm_dps], thy'') = Global_Theory.add_defs false
       [((Binding.map_name (suffix "_def") binding, def), [])] thy'
    val thy_' = Sign.add_consts [(binding', trsT, NoSyn)] thy''
    val Rf = Sign.full_name thy_' binding'
    val def' = Logic.mk_equals (Const (Rf, trsT), trs_dps)
    val ([defthm], thy_'') = Global_Theory.add_defs false
       [((Binding.map_name (suffix "_def") binding', def'), [])] thy_'
    val ctxt'' = Proof_Context.init_global thy_''

    val inclusions = extract_subsystems ctxt'' defthm
    fun rules f = let
      val rls_trees = (snd o the) (List.find (fn e => fst e = f) dp_rules)
      val thms = maps (map (fn (lr,t) => (extract_dp_rule ctxt'' lr defthm_dps,t))) rls_trees
      in (sharp_prefix ^ f, thms) end
  in thy_''
    |> map_dp_deftab (fold (fn f => Symtab.update (f, (Pf, defthm_dps, Rf, defthm))) fs) 
    |> map_dp_treetab (K treetab')
    |> map_dp_ruletab (fold Symtab.update_new (map rules (fs @ case_syms)))
    |> map_incltab (fold Symreltab.update_new inclusions) 
  end

fun declare_constrs tname thy =
let
  val constrs = the (Datatype.get_constrs thy tname)
in
  add_constrs (map fst constrs) thy
end

fun all_trs_defs context = (get_deftab (Context.theory_of context)
       |> Symtab.dest |> map snd |> AList.group (op =) |> map (hd o snd)) @
       (get_dp_deftab (Context.theory_of context)
       |> Symtab.dest |> map snd |> maps (fn (f,t,g,u) => [(f,t),(g,u)]) |> AList.group (op =) |> map (hd o snd))


val setup = Embedding.setup
  #> Subst.setup
  #> Global_Theory.add_thms_dynamic (Binding.name "sim", map snd o Symtab.dest o Simulation.get)
  #> Global_Theory.add_thms_dynamic (Binding.name "all_trs_defs", all_trs_defs)

end
